library(raster)
library(rasterVis)
library(rgdal)
library(ncdf4)
library(ncdf.tools)
library(ncdf4.helpers)
library(data.table)
library(magrittr)
#library(doParallel)
#library(foreach)
library(eurocordexr)
library(PCICt)

#registerDoParallel(parallel::detectCores() - 1)

# testing bias-correction methods for tas
BA.method<-"QM" # start with Empirical Quantile Mapping
# the QM implementation needs to be improved by following https://github.com/SvenKotlarski/qmCH2018
path_in_mod<-"/../CORDEX/" #already remapped and cropped files 
variable<-"tasmin"
inventory<-readRDS(paste0(path_in_mod,"/",variable,"_inventory.RDS")) #or a list for the selected models
CORDEX.models<-unique(paste0(inventory$gcm,"+",inventory$institute_rcm)) # total number of used models
hist_years <- 1950:2005 # years in historical period
proj_years <- 2006:2100 # years in projections
cal_years  <- 1981:2010 # years for the calibration

# step 1
# load daily observations for the variable under evaluation and for the reference period 
# observations are split by month and stored in separate folders for each year
ref.years<-1981:2010
path_in_obs<-"/../OBS_DATA/"
list<-list.files(paste0(path_in_obs,variable,"/",ref.years[1]),pattern=".nc",full.names = T)
obs.stack<-stack(lapply(list,stack))
 for(i in ref.years[-1]){
   list<-list.files(paste0(path_in_obs,variable,"/",i),pattern=".nc",full.names = T)
   suppressWarnings(tmp<-stack(lapply(list,stack)))
   obs.stack<-stack(obs.stack,tmp)
 }

idx.TSTcell<-which(!is.na(getValues(obs.stack[[1]]))) # indices of cells which will be processed
scenario<-"rcp85"
# convert stack into matrix with only points of interest 
# row=time col=points
obs.dat<-as.data.frame(obs.stack)
obs.dat<-obs.dat[idx.TSTcell,]; obs.dat<-t(data.matrix(obs.dat))

# step 2 
# load CORDEX for the specific simulation and run the QM correction
dates.ref<-seq(as.Date(paste0(min(cal_years),"/01/01")),as.Date(paste0(max(cal_years),"/12/31")),"day")
dates.hist.full<-seq(as.Date(paste0(min(hist_years),"/01/01")),as.Date(paste0(max(hist_years),"/12/31"),"day")) # CORDEX historical period
dates.proj.full<-seq(as.Date(paste0(min(proj_years),"/01/01")),as.Date(paste0(max(proj_years),"/01/01")),"day") # CORDEX scenario period
dates.full<-c(dates.hist.full,dates.proj.full) # full analysed period

# loop over each gcm+rcm simulation --> this loop should be parallelized
  for(i.mod in seq_along(CORDEX.models)){ 
    
    i.gcm<-strsplit(CORDEX.models[i.mod],"+",fixed=T)[1]
    i.rcm<-strsplit(CORDEX.models[i.mod],"+",fixed=T)[2]
    
    #load full data series for the gcm+rcm simulation
    dat.mod<-load_CORDEX(gcm=i.gcm,rcm=i.rcm,
                         path_in_mod=path_in_mod, variable=variable,
                         cells=idx.TSTcell,
                         dates.historical=dates.hist.full,
                         dates.proections=dates.proj.full)

    # matrix for storing QM corrected simulations
    # row=time, col=points
    mod.qm<-matrix(ncol=length(idx.TSTcell),nrow=length(dates.full))
    
    
    # the loop is currently on months but it should run over calendar days 
    # and a 91-day window centered on the day should be selected to extract the correction: 
    # see implementation here: https://github.com/SvenKotlarski/qmCH2018 
    for (i.month in 1:12){ 
      
      # extract observations in the considered month
      obs<-obs.dat[which(month(dates.ref)==i.month),]
      dates.obs<-dates.ref[which(month(dates.ref)==i.month)]
      # extract model data over the calibration period for the monthly window
      mod.hist<-dat.mod[which(year(dates.full)%in%cal_years & month(dates.full)==i.month),]
      
      for(i.cell in seq_along(idx.TSTcell)){
        # fit quantiles and extract transfer function over the calibration period
        # wet.day=T for precipitation only
        # this code should be modified in order to change the quantiles used
        # currently the function considers all centiles from 0 to 1 
        # it would be better to use 0.01 to 0.99 only and constant correction for outside this range
        # it would be more robust when dealing with extremes --> qm with custom quantiles here:
        # https://github.com/SvenKotlarski/qmCH2018
        qm<-fitQmapQUANT(obs=as.numeric(obs[,i.cell]),mod=mod.hist[,i.cell],wet.day = F)
        # apply quantile mapping to the model simulation over the entire spanned period e.g. 1950-2100
        mod.qm[which(month(dates.full)==i.month),i.cell]<-doQmapQUANT(mod.dat[which(month(dates.full)==i.month),i.cell],qm)     
      }
    }
    # save as .rda (rdata file) but it has to be exported as netCDF
    name.out<-paste0(inventory[which(inventory$gcm==i.gcm & inventory$institute_rcm==i.rcm & 
                                       inventory$experiment==scenario),-c(9,10)],collapse="_")
    save(mod.qm,file=paste0(path_out,variable,"/",BA.method,"/BiasAjdust_",name.out,".rda"),compress=T)
  }