# imports
library(raster)
library(rasterVis)
library(rgdal)
library(ncdf4)
library(ncdf4.helpers)
library(data.table)
library(magrittr)
library(PCICt)

# load local modules
source('./config.R')
source('./02_load_CORDEX_function.R')
source("./03a_QM_DOY_utils.R")

# TODO: parallelization of different models
# registerDoParallel(parallel::detectCores() - 1)

# ------------------------------------------------------------------------------
# Step 1: load daily observations for the calibration period -------------------
# ------------------------------------------------------------------------------

# search observations in defined path: observations are split by month and
# stored in separate folders for each year
obs <- list.files(paste0(obs_path, '/', variable, '/', cal_years[1]),
                         pattern=".nc", full.names = T)

# create raster stack
obs.stack <- stack(lapply(obs, stack))

# read all years of the calibration period
for(i in cal_years[-1]){
  list <- list.files(paste0(obs_path, '/', variable, "/", i),
                     pattern=".nc", full.names = T)
  tmp <- stack(lapply(list, stack))
  obs.stack <- stack(obs.stack, tmp)
}

# indices of cells which will be processed
idx.TSTcell <- which(!is.na(getValues(obs.stack[[1]])))

# convert stack into matrix with only points of interest: (time, points)
obs.dat <- as.data.frame(obs.stack)
obs.dat <- obs.dat[idx.TSTcell,]
obs.dat<-t(data.matrix(obs.dat))

# ------------------------------------------------------------------------------
# Step 2: define daily time periods --------------------------------------------
# ------------------------------------------------------------------------------

# calibration period
dates.ref <- seq(as.Date(paste0(min(cal_years), "/01/01")),
                 as.Date(paste0(max(cal_years), "/12/31")),
                 "day")

# historical period
dates.hist.full<-seq(as.Date(paste0(min(hist_years), "/01/01")),
                     as.Date(paste0(max(hist_years), "/12/31")),
                             "day")

# projected period
dates.proj.full<-seq(as.Date(paste0(min(proj_years), "/01/01")),
                     as.Date(paste0(max(proj_years), "/01/01")),
                     "day")

# full period
dates.full<- c(dates.hist.full, dates.proj.full)

# loop over each gcm+rcm simulation
for(i.mod in seq_along(CORDEX.models)){ 
  
  i.gcm<-strsplit(CORDEX.models[i.mod], "+", fixed=T)[1]
  i.rcm<-strsplit(CORDEX.models[i.mod], "+", fixed=T)[2]
  
  # load full data series for the gcm + rcm simulation
  dat.mod<-load_CORDEX(gcm=i.gcm,rcm=i.rcm,
                       path_in_mod=mod_path, variable=variable,
                       cells=idx.TSTcell,
                       dates.historical=dates.hist.full,
                       dates.proections=dates.proj.full)
  
  # matrix for storing QM corrected simulations
  # row=time, col=points
  qm.mat<-matrix(ncol=length(idx.TSTcell), nrow=length(dates.full))
  
  for(i.cell in 1:length(idx.TSTcell)){
    # load the series for the test point
    series.mod.cal <- dat.mod[which(year(dates.full) %in% cal_years), i.cell]
    series.obs.cal <- as.numeric(obs.dat.conv[which(year(dates.ref)
                                                    %in% cal_years), i.cell])
    series.mod.all <- dat.mod[which(year(dates.full) %in% cor_years), i.cell]
    
    # set to zero pr corresponding to dry days
    if ((variable=='pr') & FA) {
      series.mod.cal[series.mod.cal < pr.wet] <- 0
      series.obs.cal[series.obs.cal < pr.wet] <- 0
      series.mod.all[series.mod.all < pr.wet] <- 0
    }
    
    # extract the DOY index for the full model series 
    # the index for 29th Feb is set to 999 in all leap years
    # 29th Feb is then handled as 1st Mar
    doy.full <- qm.doystring(min(cor_years), max(cor_years))
    
    # the quantile function for each DOY for both model and observation data
    c_mod <- qm.cdf.doy(series.mod.cal, min(cal_years), max(cal_years),
                        wday.width, minq, maxq, incq, variable, pr.wet)
    c_obs <- qm.cdf.doy(series.obs.cal, min(cal_years), max(cal_years),
                        wday.width, minq, maxq, incq, variable, pr.wet)
    
    # here the correction for each quantile and DOY (additive) is computed
    if((length(c_mod$cdf.matrix[1,]) != length(c_obs$cdf.matrix[1,])) &
       (length(c_mod$cdf.matrix[,1]) != length(c_obs$cdf.matrix[,1]))){
      print("ERROR: doy-based CDF of model and observations do not have the same dimension")
    } else {
      # correction factor for each DOY and quantile
      dif.cdf	<- c_mod$cdf.matrix - c_obs$cdf.matrix
    }
    
    if(variable == "pr"){
      # apply the quantile correction by adding an adjustment of precipitation
      # frequency in model data
      cc <- qm.empqm.doy.fa(series.mod.all, doy.full, c_mod$cdf.matrix,
                            c_obs$cdf.matrix, c_obs$cdf.wet, dif.cdf,
                            c_obs$f.wet, c_mod$f.wet, pr.wet,
                            method=qm.method, variable, minq, maxq, incq)
      
      # check if negative corrected precipitation occur and set to 0
      cc$qm.corrected.series[cc$qm.corrected.series < pr.wet] <- 0
      
    } else {
      # quantile mapping
      cc <- qm.empqm.doy(series.mod.all.in, doy.full, c_mod$cdf.matrix, dif.cdf, 
                         method=qm.method, variable, minq, maxq, incq)
    }
    
    # store the corrected series in the output matrix
    qm.mat[,i.cell]<-cc$qm.corrected.series
  } # close loop on grid cell
} # close loop on models